<?php declare(strict_types=1);

namespace MyPrayerJournal\Domain;

use BitBadger\PDODocument\{Custom, DocumentException, DocumentList, Find, Mapper\DocumentMapper};
use DateTimeImmutable;
use Exception;
use JsonSerializable;
use MyPrayerJournal\Table;
use Visus\Cuid2\Cuid2;

/**
 * A prayer request
 */
class Request implements JsonSerializable
{
    /**
     * @param string $id The ID for the request
     * @param string $enteredOn The date/time this request was originally entered
     * @param string $userId The ID of the user to whom this request belongs
     * @param string|null $snoozedUntil The date/time the snooze expires for this request (null = not snoozed)
     * @param string|null $showAfter The date/time the current recurrence period is over (null = immediate)
     * @param Recurrence $recurrence The recurrence for this request
     * @param History[] $history The history of this request
     * @param Note[] $notes Notes regarding this request
     * @throws Exception If the ID generation fails
     */
    public function __construct(public string $id = '', public string $enteredOn = '', public string $userId = '',
                                public ?string $snoozedUntil = null, public ?string $showAfter = null,
                                public Recurrence $recurrence = new Recurrence(RecurrencePeriod::Immediate),
                                public array $history = [], public array $notes = [])
    {
        if ($id == '') {
            $this->id = (new Cuid2())->toString();
        }
    }

    /**
     * Get the current text for this request
     *
     * @return string The most recent text for the request
     */
    public function currentText(): string
    {
        foreach ($this->history as $hist) if (isset($hist->text)) return $hist->text;
        return '';
    }

    /**
     * Get the date/time this request was last marked as prayed
     *
     * @return string|null The date/time this request was last marked as prayed
     */
    public function lastPrayed(): ?string
    {
        foreach ($this->history as $hist) if ($hist->action == RequestAction::Prayed) return $hist->asOf;
        return null;
    }

    /**
     * Has this request been answered?
     *
     * @return bool True if the request is answered, false if not
     */
    public function isAnswered(): bool
    {
        return $this->history[0]->action == RequestAction::Answered;
    }

    /**
     * Is this request currently snoozed?
     *
     * @return bool True if the request is snoozed, false if not
     * @throws Exception If the snoozed until date/time is not valid
     */
    public function isSnoozed(): bool
    {
        return isset($this->snoozedUntil) && new DateTimeImmutable($this->snoozedUntil) > new DateTimeImmutable('now');
    }

    /**
     * Is this request currently not shown due to recurrence?
     *
     * @return bool True if the request is pending, false if not
     * @throws Exception If the snoozed or show-after date/times are not valid
     */
    public function isPending(): bool
    {
        return !$this->isSnoozed()
            && isset($this->showAfter)
            && new DateTimeImmutable($this->showAfter) > new DateTimeImmutable('now');
    }

    public function jsonSerialize(): mixed
    {
        $values = [
            'id'         => $this->id,
            'enteredOn'  => $this->enteredOn,
            'userId'     => $this->userId,
            'recurrence' => $this->recurrence,
            'history'    => $this->history,
            'notes'      => $this->notes
        ];
        if (isset($this->snoozedUntil)) $values['snoozedUntil'] = $this->snoozedUntil;
        if (isset($this->showAfter))    $values['showAfter']    = $this->showAfter;
        return $values;
    }

    /**
     * Find a request by its ID
     *
     * @param string $id The ID of the request
     * @return Request|false The request if it is found and belongs to the current user, false if not
     * @throws DocumentException If any is encountered
     */
    public static function byId(string $id): Request|false
    {
        $req = Find::byId(Table::REQUEST, $id, self::class);
        return ($req && $req->userId == $_SESSION['user_id']) ? $req : false;
    }

    /**
     * Get the user's active journal requests
     *
     * @return DocumentList<Request> The requests for the user's journal
     * @throws DocumentException If any is encountered
     */
    public static function forJournal(): DocumentList
    {
        $table = Table::REQUEST;
        return Custom::list(<<<SQL
            SELECT data, (
                   SELECT h.value->>'asOf' as_of
                     FROM $table i LEFT JOIN json_each(i.data, '$.history') h
                    WHERE r.data->>'id' = i.data->>'id' AND h.value->>'action' = 'Prayed'
                    LIMIT 1) last_prayed
              FROM $table r
             WHERE data->>'userId'               = :userId
               AND data->>'$.history[0].action' <> 'Answered'
               AND (data->>'snoozedUntil' IS NULL OR data->>'snoozedUntil' < datetime('now'))
               AND (data->>'showAfter'    IS NULL OR data->>'showAfter'    < datetime('now'))
             ORDER BY coalesce(last_prayed, data->>'snoozedUntil', data->>'showAfter', data->>'$.history[0].asOf')
            SQL, [':userId' => $_SESSION['user_id']], new DocumentMapper(self::class));
    }

    /**
     * Get either the user's active or answered requests
     *
     * @param bool $active True to retrieve active requests, false to retrieve answered requests
     * @param bool $snoozed True to retrieve only snoozed requests
     * @return DocumentList<Request> The requests matching the criteria
     * @throws DocumentException If any is encountered
     */
    private static function forUser(bool $active = true, bool $snoozed = false): DocumentList
    {
        $table = Table::REQUEST;
        $op    = $active ? '<>' : '=';
        $extra = $snoozed ? "AND datetime(data->>'snoozedUntil') > datetime('now')" : '';
        $order = $active
            ? "coalesce(data->>'snoozedUntil', data->>'showAfter', last_prayed, data->>'$.history[0].asOf')"
            : "data->>'$.history[0].asOf' DESC";
        return Custom::list(<<<SQL
            SELECT data, (
                   SELECT h.value->>'asOf' as_of
                     FROM $table i LEFT JOIN json_each(i.data, '$.history') h
                    WHERE r.data->>'id' = i.data->>'id' AND h.value->>'action' = 'Prayed'
                    LIMIT 1) last_prayed
              FROM $table r
             WHERE data->>'userId'                = :userId
               AND data->>'$.history[0].action' $op 'Answered' $extra
             ORDER BY $order
            SQL, [':userId' => $_SESSION['user_id']], new DocumentMapper(self::class));
    }

    /**
     * Get a list of active requests for a user
     *
     * @return DocumentList<Request> The user's active requests
     * @throws DocumentException If any is encountered
     */
    public static function active(): DocumentList
    {
        return self::forUser();
    }

    /**
     * Get a list of answered requests for a user
     *
     * @return DocumentList<Request> The user's answered requests
     * @throws DocumentException If any is encountered
     */
    public static function answered(): DocumentList
    {
        return self::forUser(false);
    }

    /**
     * Get a list of snoozed requests for a user
     *
     * @return DocumentList<Request> The user's snoozed requests
     * @throws DocumentException If any is encountered
     */
    public static function snoozed(): DocumentList
    {
        return self::forUser(snoozed: true);
    }
}
