<?php declare(strict_types=1);

namespace MyPrayerJournal\UI;

use DateTimeImmutable;
use Exception;

/**
 * A type of relative date, along with the formatting strings
 */
enum RelativeDate: string
{
    case LessThanXMinutes = 'less than a minute|less than %d minutes';
    case XMinutes = 'a minute|%d minutes';
    case AboutXHours = 'about an hour|about %d hours';
    case XHours = 'an hour|%d hours';
    case XDays = 'a day|%d days';
    case AboutXWeeks = 'about a week|about %d weeks';
    case XWeeks = 'a week|%d weeks';
    case AboutXMonths = 'about a month|about %d months';
    case XMonths = 'a month|%d months';
    case AboutXYears = 'about a year|about %d years';
    case XYears = 'a year|%d years';
    case OverXYears = 'over a year|over %d years';
    case AlmostXYears = 'almost a year|almost %d years';

    //  Many thanks to date-fns (https://date-fns.org) for this logic
    /**
     * Format the distance between two dates
     *
     * @param string|DateTimeImmutable $from The starting date/time
     * @param string|DateTimeImmutable $to The ending date/time
     * @return string The distance between two dates
     * @throws Exception If date/time objects cannot be created
     */
    public static function between(string|DateTimeImmutable $from, string|DateTimeImmutable $to): string
    {
        $aDay        =  1_440.0;
        $almost2Days =  2_520.0;
        $aMonth      = 43_200.0;
        $twoMonths   = 86_400.0;

        $dtFrom  = is_string($from) ? new DateTimeImmutable($from) : $from;
        $dtTo    = is_string($to)   ? new DateTimeImmutable($to)   : $to;
        $minutes = abs($dtFrom->getTimestamp() - $dtTo->getTimestamp()) / 60;
        $months  = round($minutes / $aMonth);
        $years   = round($months / 12);

        [$type, $number] = match (true) {
            $minutes < 1.0          => [RelativeDate::LessThanXMinutes, 1],
            $minutes < 45.0         => [RelativeDate::XMinutes,         round($minutes)],
            $minutes < 90.0         => [RelativeDate::AboutXHours,      1],
            $minutes < $aDay        => [RelativeDate::AboutXHours,      round($minutes / 60)],
            $minutes < $almost2Days => [RelativeDate::XDays,            1],
            $minutes < $aMonth      => [RelativeDate::XDays,            round($minutes / $aDay)],
            $minutes < $twoMonths   => [RelativeDate::AboutXMonths,     round($minutes / $aMonth)],
            $months      < 12       => [RelativeDate::XMonths,          round($minutes / $aMonth)],
            $months % 12 < 3        => [RelativeDate::AboutXYears,      $years],
            $months % 12 < 9        => [RelativeDate::OverXYears,       $years],
            default                 => [RelativeDate::AlmostXYears,     $years]
        };
        [$singular, $plural] = explode('|', $type->value);
        $value = $number == 1 ? $singular : sprintf($plural, $number);
        return $dtFrom > $dtTo ? "$value ago" : "in $value";
    }
}
